# [Valid version numbers](https://packaging.python.org/en/latest/discussions/versioning/#versioning)
# [PEP 440 – Version Identification and Dependency Specification](https://peps.python.org/pep-0440/)
# [Specifying Your Project’s Version](https://setuptools.pypa.io/en/latest/userguide/distribution.html)
# [N!]N(.N)*[{a|b|rc}N][.postN][.devN]
VERSION = "0.1.0a2"
